/**
   BLE2Wifi_ESP8266Web_example.ino
   by Matthew Ford,  2022/04/23
   (c)2022 Forward Computing and Control Pty. Ltd.
   NSW, Australia  www.forward.com.au
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.
*/

// here only the temperature reading is displayed, but all the values are parsed

// This example is for ESP8266 boards --  Adafruit ESP8266 HUZZAH in this example
// It is the WiFi side of a BLE2WiFi bridge for BLE very low power temperature sensors
// see https://www.forward.com.au/pfod/BLE/LowPower_2022/Temp/index.html
// Configure a staticIP so you can connect the the temperature server

// Sensor advertised names start with
//  T_.. for Temp(C) sensors, H_.. for Temp(C),Humditiy(%) sensors, W_.. for Temp(C),Humdity(%),Barometer(hPa) sensors, WL_.. for Temp(C),Humdity(%),Barometer(hPa),Lux level(Lux)
//  keep this LOCAL_NAME short as it uses up available advertising data space (see below)
// Format of advertized name
// For T_.. devices   LOCAL_NAME,degC
// For H_.. devices   LOCAL_NAME,degC,%RH
// For W_.. devices   LOCAL_NAME,degC,%RH,hPa
// For WL_.. devices  LOCAL_NAME,degC,%RH,hPa,Lux

#include "ESPAutoWiFiConfig.h" // download ESPAutoWiFiConfig.zip from  https://www.forward.com.au/pfod/ESPAutoWiFiConfig/index.html
// includes ESP8266BufferedClient
#include <SafeString.h>  // install the SafeString library via the Arduino Library manager
#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include "LastDataSeen.h"
#include "pfodLinkedPointerList.h"

ESP8266WebServer server(80);

// settings for the Auto WiFi connect
int ledPin = 0; // for Adafruit ESP8266 HUZZAH - onboard Led connected to GPIO0
// use ledPin = 2 for ESP-01S (Blue led on GPIO2)
// use ledPin = 1 for ESP-01 (Blue led on TX (GPIO1), ESP-01 also has a Red power led)
bool highForLedOn = false; // need to make output low to turn Adafruit ESP8266 HUZZAH's GPIO0 Led ON
size_t eepromOffset = 0; // if you use EEPROM.begin(size) in your code add the size here so AutoWiFi data is written after your data

//#define DEBUG

void startWebServer();

cSF(sensorData, 50); // max name size is 32 so 50 is enough
const size_t MAX_BLE_NAME_SIZE = 32;
char T_DEVICE[] = "T_"; // T_ devices have temp only (in degC)
char H_DEVICE[] = "H_"; // H_ devices have temp(degC),RH(%)
char W_DEVICE[] = "W_"; // W_ devices have temp(degC),RH(%),AirPressure(hPa)
char WL_DEVICE[] = "WL_"; // WL_ devices have temp(degC),RH(%),AirPressure(hPa),Lux
cSF(sfDevice, MAX_BLE_NAME_SIZE);
cSF(sf_TempC, MAX_BLE_NAME_SIZE);
cSF(sf_RH, MAX_BLE_NAME_SIZE);
cSF(sf_hPa, MAX_BLE_NAME_SIZE);
cSF(sf_lx, MAX_BLE_NAME_SIZE);
float tempC, RH, hPa, lx;

pfodLinkedPointerList<LastDataSeen> listOfLastDataSeen;

void setup() {
  // setup code that must be run every time
  Serial.begin(115200);
  Serial.println();
#ifdef DEBUG
  for (int i = 10; i > 0; i--) {
    Serial.print(i); Serial.print(' ');
    delay(500);
  }
  Serial.println();
#endif
  SafeString::setOutput(Serial);

#ifdef DEBUG
  setESPAutoWiFiConfigDebugOut(Serial); // turns on debug output for the ESPAutoWiFiConfig code
#endif

  if (ESPAutoWiFiConfigSetup(ledPin, highForLedOn, eepromOffset)) { // check if we should start access point to configure WiFi settings
    return; // in config mode so skip rest of setup
  }

  // when we get here the WiFi has connected !!
  // normal setup code that is run when not configuring the WiFi settings
  startWebServer();
}

bool checkValidName(SafeString &sfName) {
  return (sfName.startsWith(T_DEVICE) || sfName.startsWith(H_DEVICE) || sfName.startsWith(W_DEVICE) || sfName.startsWith(WL_DEVICE));
}

bool decode(SafeString &sfName) {
  sfName.trim();
  if (!checkValidName(sfName)) {
    return false;
  }
  // no need to clear the previous tokens as nextToken does that
  if (sfName.startsWith(T_DEVICE)) {
    // skip first field and pick up temp
    sfName.nextToken(sfDevice, ','); // remove name
    sfName.nextToken(sf_TempC, ','); // pick up temp
#ifdef DEBUG
    Serial.print(sfDevice); Serial.print(" Temperature:"); Serial.print(sf_TempC); Serial.println("C");
#endif
    // convert Temp
    if (!sf_TempC.toFloat(tempC)) {
      return false; // last tempC unchanged
    }
  } else if (sfName.startsWith(H_DEVICE)) {
    // skip first field and pick up temp
    sfName.nextToken(sfDevice, ','); // remove name
    sfName.nextToken(sf_TempC, ','); // pickup Temp
    sfName.nextToken(sf_RH, ','); // pickup RH
#ifdef DEBUG
    Serial.print(sfDevice);
    Serial.print(" Temperature:"); Serial.print(sf_TempC); Serial.print("C");
    Serial.print(" RH:"); Serial.print(sf_RH); Serial.print("%");
    Serial.println();
#endif
    if (!sf_TempC.toFloat(tempC)) {
      return false; // not a valid float, last tempC unchanged
    }
    if (!sf_RH.toFloat(RH)) {
      return false; // not a valid float, last RH unchanged
    }
  } else if (sfName.startsWith(W_DEVICE)) {
    // skip first field and pick up temp
    sfName.nextToken(sfDevice, ','); // remove name
    sfName.nextToken(sf_TempC, ','); // pickup Temp
    sfName.nextToken(sf_RH, ','); // pickup RH
    sfName.nextToken(sf_hPa, ','); // pickup Air Pressure
#ifdef DEBUG
    Serial.print(sfDevice);
    Serial.print(" Temperature:"); Serial.print(sf_TempC); Serial.print("C");
    Serial.print(" RH:"); Serial.print(sf_RH); Serial.print("%");
    Serial.print(" Barometer:"); Serial.print(sf_hPa); Serial.print(" hPa");
    Serial.println();
#endif
    if (!sf_TempC.toFloat(tempC)) {
      return false; // not a valid float, last tempC unchanged
    }
    if (!sf_RH.toFloat(RH)) {
      return false; // not a valid float, last RH unchanged
    }
    if (!sf_hPa.toFloat(hPa)) {
      return false; // not a valid float, last hPA unchanged
    }
  } else if (sfName.startsWith(W_DEVICE)) {
    // skip first field and pick up temp
    sfName.nextToken(sfDevice, ','); // remove name
    sfName.nextToken(sf_TempC, ','); // pickup Temp
    sfName.nextToken(sf_RH, ','); // pickup RH
    sfName.nextToken(sf_hPa, ','); // pickup Air Pressure
    sfName.nextToken(sf_lx, ','); // pickup Air Pressure
#ifdef DEBUG
    Serial.print(sfDevice);
    Serial.print(" Temperature:"); Serial.print(sf_TempC); Serial.print("C");
    Serial.print(" RH:"); Serial.print(sf_RH); Serial.print("%");
    Serial.print(" Barometer:"); Serial.print(sf_hPa); Serial.print(" hPa");
    Serial.print(" Lux:"); Serial.print(sf_lx); Serial.print(" lx");
    Serial.println();
#endif
    if (!sf_TempC.toFloat(tempC)) {
      return false; // not a valid float,  last tempC unchanged
    }
    if (!sf_RH.toFloat(RH)) {
      return false; // not a valid float,  last RH unchanged
    }
    if (!sf_hPa.toFloat(hPa)) {
      return false; // not a valid float,  last hPA unchanged
    }
    if (!sf_lx.toFloat(lx)) {
      return false; // not a valid float,  last lx unchanged
    }
  } else {
#ifdef DEBUG
    Serial.print("Error: Device not supported -- "); Serial.println(sfName);
#endif
    return false;
  }
  return true;
}

// returns NULL if not found
LastDataSeen* getLastSeen(SafeString &name) {
  LastDataSeen *devicePtr = listOfLastDataSeen.getFirst();
  while (devicePtr) {
    if (name.startsWith(devicePtr->getDeviceName())) {
      return devicePtr;
    } // else
    devicePtr = listOfLastDataSeen.getNext();
  }
  return devicePtr; // null not found
}

void updateLastDataSeen() {
  LastDataSeen *devicePtr = getLastSeen(sfDevice);
  if (devicePtr == NULL) {
    // not found add this new one
    devicePtr = new LastDataSeen(sfDevice.c_str());
    listOfLastDataSeen.add(devicePtr);
  }
  // update it
  devicePtr->setTemp(tempC);
}


void loop() {
  //  .. other code that MUST run all the time
  if (ESPAutoWiFiConfigLoop()) {  // handle WiFi config webpages
    return;  // skip the rest of the loop until config finished
  }

  // .. normal loop code here that is run when not configuring the WiFi settings
  // check Serial for new data

  if (sensorData.readUntil(Serial, '\n')) {
    // have a new line or sensorData filled up
    sensorData.trim();
    if (sensorData.length() > MAX_BLE_NAME_SIZE) {
      // not advertising data just echo
#ifdef DEBUG
      Serial.print("Input not sensor data -- "); Serial.println(sensorData);
#endif
    } else {
      if (checkValidName(sensorData)) {
        if (decode(sensorData)) {
          updateLastDataSeen(); // uses the gobals
        }
      }
    }
    sensorData.clear(); // for next read
  }
  server.handleClient();
}

// ========== trivial webserver example =============
void notFound() {
  server.send(404, "text/plain", "Not found");
}

void handleRoot() {
  String msg =
    "<html>\
  <head>\
    <meta http-equiv='refresh' content='5'/>\
    <title>BLE Temperature Sensors</title>\
    <style>\
      body { background-color: #cccccc; font-family: Arial, Helvetica, Sans-Serif; Color: #000088; }\
    </style>\
  </head>\
  <body>";
  msg += "The temperatures are:-<br>";
  // loop through and add reading found
  LastDataSeen *devicePtr = listOfLastDataSeen.getFirst();
  if (devicePtr == NULL) {
    msg += "No devices found so far.<br>";
  } else {
    while (devicePtr != NULL) {
      msg += "<h1>";
      msg += devicePtr->getDeviceName();
      msg += "  Temperature: ";
      msg += devicePtr->getTemp();
      msg += "&degC  ";
      msg += "<font size=\"-1\">";
      msg += (millis() - devicePtr->getLastTimeDataSeen()) / 1000.0;
      msg += " sec ago</font>";
      msg += "</h1>";
      devicePtr = listOfLastDataSeen.getNext();
    }
  }
  msg += "</body></html>";

  server.send(200, "text/html", msg);
}

void startWebServer() {
  server.on("/", handleRoot);
  server.onNotFound(notFound);
  server.begin();
}
